/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This is for the Amiga prlink 8/8-bit cable.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <exec/exec.h>
#include <resources/misc.h>
#include <hardware/cia.h>
#include <clib/exec_protos.h>
#include <clib/misc_protos.h>

#define __PRAMIGA88_C__
#include "prtrans.h"

#define DEBOUNCE	1	/* debounce the ack line */
#define DEBOUNCE2	1	/* debounce the ack line more thoroughly */
#define Debug(x)		/* nothing, without debugging */

/*
 * The CIAs are volatile, otherwise the while loop waiting for ACK to
 * change is "optimized".
 */
extern volatile struct CIA ciaa;
extern volatile struct CIA ciab;

struct Library *MiscBase;
static unsigned long MiscAllocated;
static unsigned char ack;
#if DEBOUNCE2
int		bounce;
#endif
static char	myname[] = __BASE_FILE__;

#define BUSY	(1 << CIAB_PRTRBUSY)	/* == 0x01 */
#define POUT	(1 << CIAB_PRTRPOUT)	/* == 0x02 */

#define DATA		ciaa.ciaprb
#define DATA_DDR	ciaa.ciaddrb
#define CTL		ciab.ciapra
#define CTL_DDR 	ciab.ciaddra

int
prinit(void)
{
    Debug(printf("prinit\n");)
    if (MiscAllocated == 0) {
	atexit(prclose);

	MiscBase = OpenResource("misc.resource");
	if (MiscBase == NULL) {
	    printf("Can't open misc.resource\n");
	    return -1;
	} {
	    const char     *user;

	    user = AllocMiscResource(MR_PARALLELPORT, myname);
	    if (user != NULL) {
		printf("Printer data lines already in use by %s.\n", user);
		return -2;
	    }
	    MiscAllocated = 1 << MR_PARALLELPORT;
	    user = AllocMiscResource(MR_PARALLELBITS, myname);
	    if (user != NULL) {
		printf("Printer control lines already in use by %s.\n", user);
		return -3;
	    }
	    MiscAllocated |= 1 << MR_PARALLELBITS;
	}

	/* turn POUT into output (BUSY is input) */
	CTL |= POUT;	/* and set it high */
	CTL_DDR = (CTL_DDR | POUT) & ~BUSY;
	DATA = 0xFF;	/* and high */
	DATA_DDR = 0xFF;	/* set all lines to output */
    }

    ack = CTL & BUSY;	/* get the current state of the BUSY line */

    return 0;
}

void
prclose(void)
{
    Debug(printf("prclose\n");)
#if DEBOUNCE2
    if (bounce != 0) {
	printf("*** bounce, bounce! (%d times)\n", bounce);
	bounce = 0;
    }
#endif
    if (MiscAllocated & (1 << MR_PARALLELBITS)) {
	CTL_DDR &= ~POUT;	/* turn POUT back into input */
	FreeMiscResource(MR_PARALLELBITS);
    }
    if (MiscAllocated & (1 << MR_PARALLELPORT)) {
	DATA = 0xFF;
	DATA_DDR = 0xFF;	/* set all data lines to output */
	FreeMiscResource(MR_PARALLELPORT);
    }
    MiscAllocated = 0;
    /* There is no CloseResource() function, really! */
}

void
output(unsigned char byte)
{
    Debug(printf("output(%02x)\n", byte);)
    psend(&byte, 1);
}

void wait_output (unsigned char byte)
{
    Debug(printf("wait_output(%02x)\n", byte);)
    DATA_DDR = 0xFF;	/* set all lines to output */

    DATA = byte;
    CTL &= ~POUT;		/* drop the -FLAG to low */
    while (ack == (BUSY & CTL)) {
	Debug(printf("sleeping...\n");)
	usleep (SLEEP_TIME); /* wait for BUSY to change */
    }
    CTL |= POUT;		/* raise -FLAG again */
    ack ^= BUSY;		/* store the new state of -BUSY */
}

unsigned
input(void)
{
    unsigned char   byte;

    Debug(printf("input() = "); fflush(stdout);)
    preceive(&byte, 1);
    Debug(printf("%02x\n", byte);)
    return byte;
}

unsigned
wait_input(void)
{
    unsigned char   byte;

    Debug(printf("wait_input() = "); fflush(stdout);)
    CTL &= ~POUT;		/* drop the -FLAG to low */
    while (ack == (BUSY & CTL)) /* wait for BUSY to change */
	usleep (SLEEP_TIME);

    preceive(&byte, 1);
    Debug(printf("%02x\n", byte);)
    return byte;
}

void send_switch() {}
void receive_switch() {}

void
psend(unsigned char *buffer, unsigned length)
{
    DATA_DDR = 0xFF;	/* set all lines to output */

#if defined(__GNUC__) && !NO_ASM
/* INDENT OFF */
    asm("
	subql #1,%1
	moveq #-1,d0
	cmpl %1,d0
	jeq .send
.sbyteloop:
	moveb %2@+,%4@
	andb %6,%5@
.sackloop:
	moveb %5@,d0
	andb %7,d0
	cmpb %3,d0
	jeq .sackloop

	orb %8,%5@
	eorb %7,%3
	dbra %1,.sbyteloop
	clrw %1
	subql #1,%1
	jcc .sbyteloop
.send:
    " : "=d" (ack)              /* 0 output */
      : "d" (length),           /* 1 input */
	"a" (buffer),           /* 2 */
	"0" (ack),              /* 3 */
	"a" (&DATA),            /* 4 */
	"a" (&CTL),             /* 5 */
	"g" (~POUT),            /* 6 */
	"g" (BUSY),             /* 7 */
	"g" (POUT)              /* 8 */
      : "d0"                    /* clobbered */
    );
    /* INDENT ON */
#else
    while (length--) {
	DATA = *buffer++;
	CTL &= ~POUT;		/* drop the -FLAG to low */
	while (ack == (BUSY & CTL));	/* wait for BUSY to change */
	CTL |= POUT;		/* raise -FLAG again */
	ack ^= BUSY;		/* store the new state of -BUSY */
    }
#endif
}

void
preceive(unsigned char *buffer, unsigned length)
{
    DATA_DDR = 0x00;		/* set the bidirectional signals to input */

#if defined(__GNUC__) && !NO_ASM
/* INDENT OFF */
    asm("
	subql #1,%1
	moveq #-1,d0
	cmpl %1,d0
	jeq .rend

.rbyteloop:
	andb %6,%5@
.rackloop:
	moveb %5@,d0
	andb %7,d0
	cmpb %3,d0
	jeq .rackloop

	orb %8,%5@
	moveb %4@,%2@+

	eorb %7,%3
	dbra %1,.rbyteloop
	clrw %1
	subql #1,%1
	jcc .rbyteloop

.rend:
    " : "=d" (ack)              /* 0 output */
      : "d" (length),           /* 1 input */
	"a" (buffer),           /* 2 */
	"0" (ack),              /* 3 */
	"a" (&DATA),            /* 4 data */
	"a" (&CTL),             /* 5 ctl */
	"g" (~POUT),            /* 6 */
	"g" (BUSY),             /* 7 */
	"g" (POUT)              /* 8 */
      : "d0"                    /* clobbered */
    );
    /* INDENT ON */
#else
    while (length--) {
	CTL &= ~POUT;			/* drop the -FLAG to low */
#if DEBOUNCE
bounced:
#endif
	while (ack == (BUSY & CTL));	/* wait for BUSY to change */
#if DEBOUNCE
	/*
	 * With my Amiga and (PET or 64), we need to debounce the
	 * ack line. With the 64 even more seriously.
	 * I'm not sure why this is...
	 */
	if (ack == (BUSY & CTL)) goto bounced;
	if (ack == (BUSY & CTL)) goto bounced;
#if DEBOUNCE > 2
	if (ack == (BUSY & CTL)) goto bounced;
	if (ack == (BUSY & CTL)) goto bounced;
#endif
#endif
	*buffer++ = DATA;		/* read the byte */
#if DEBOUNCE2
	/* debounce */
	{
	    int 	    bouncecount;
	    unsigned char   newbyte;

	    for (bouncecount = 0; bouncecount < 3; bouncecount++) {
		newbyte = DATA;
		if (ack == (BUSY & CTL) || buffer[-1] != newbyte) {
		    buffer--;
		    bounce++;
		    goto bounced;
		}
	    }
	}
#endif
	CTL |= POUT;			/* raise -FLAG again */
	ack ^= BUSY;			/* store the new state of -BUSY */
    }
#endif
}

