/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>

#define BUFSIZE 8194
#include "prtrans.h"

#if !defined(__MAIN_C__)
#define main    main_prcart
#endif

unsigned char auxcode[] = {
/* this code starts the cartridge */
  0xa9, 0x7f,       /* lda #$7f  ; initialize the VIAs */
  0x8d, 0x1e, 0x90, /* sta $901e */
  0x8d, 0x1d, 0x90, /* sta $901d */
  0x8d, 0x2e, 0x90, /* sta $902e */
  0x8d, 0x2d, 0x90, /* sta $902d */
  0xa9, 0x00,       /* lda #$00 */
  0xa2, 0x0c,       /* ldx #$0c */
  0x9d, 0x10, 0x90, /* sta $9010,x */
  0x9d, 0x20, 0x90, /* sta $9020,x */
  0xca,             /* dex */
  0x10, 0xf7,       /* bpl . - 7 */
/* ALTJUMP: */
  0x78,             /* sei */
  0xa2, 0xff,       /* ldx #$ff */
  0x9a,             /* txs */
  0xd8,             /* cld */
  0xe8,             /* inx */
  0xa9, 0x41,       /* lda #$41 */
  0x8d, 0x04, 0xa0, /* sta $a004 (validate the boot signature) */
/* auxstart - 1: */
  0x4c, 0, 0,       /* jmp $0000 (placeholder for original cartridge start) */
/* AUXJUMP: (executed when the game is reset) */
  0xa2, 0x0d,       /* ldx #$0d */
  0x8d, 0x04, 0xa0, /* stx $a004 (invalidate the cartridge boot signature) */
  0x20, 0x8d, 0xfd, /* jsr $fd8d (copied from the reset routine) */
  0x20, 0x52, 0xfd, /* jsr $fd52 */
  0x20, 0xf9, 0xfd, /* jsr $fdf9 */
/* srvstart - 1: */
  0x20, 0, 0,       /* jsr $0000 (placeholder for prserver install routine) */
  0x4c, 0x38, 0xfd  /* jmp $fd38 */
};

unsigned char *auxstart = &auxcode[39];
unsigned char *srvstart = &auxcode[56];

#define AUXADDR 20000
#define STR(s) #s
#define AUXJUMP (auxaddr + 41)
#define ALTJUMP (auxaddr + 27)

int main (int argc, char **argv);

int main (int argc, char **argv) {
  unsigned auxaddr = AUXADDR, driveraddr, basicaddr;
  FILE *file = stdin;
  unsigned start = 0xa000, jumpaddress = 0,
           flag = 0, length;
  char **parameters = argv;
  baseaddr = portaddr[DEFAULT_PORT];

  while (*++parameters && **parameters == '-') { /* check for options */
    if (parameters[0][1] == '-') { /* "--" ends options */
      parameters++;
      break;
    }

    switch (parameters[0][1]) {
    case 'p':
      baseaddr = strtoul (*++parameters, NULL, 16);

      if (baseaddr > 3) {
        fprintf (stderr, "%s: The printer port number must be between 0 and 3.\n",
                         *argv);
        return 1;
      }

      baseaddr = portaddr[baseaddr];

      break;

    case 'a':
      if (*++parameters)
        if (65536 > (auxaddr = strtoul (*parameters, NULL, 10)))
          break;

      goto Usage;

    case 'j':
      if (*++parameters) {
        flag = 1;

        jumpaddress = strtoul (*parameters, NULL, 16);

        if (jumpaddress < 65536)
          break;
      }
      /* bleed through */

    case '?':
    case 'h':
    Usage:
      fprintf (stderr, "%s: Uploads a cartridge to the VIC-20.\n\n", *argv);
      fprintf (stderr, "Usage: %s [options] [filename.a000 [filename.6000]]\n",
                       *argv);
      fprintf (stderr, "Options:\n\t"
                         "-j addr\t"
                             "Jumps to the specified address after the transfers.\n\t\t"
                             "Not useful with auto-start cartridges.\n\t"
                         "-a addr\t"
                             "Specifies the restart address of the game.  Default=" STR(AUXADDR) ".\n\t\t"
                             "Only useful with auto-start cartridges.\n\t"
                         "-p port\t"
                             "Specify the printer port number (0 to 3).\n");
      return 1;

    default:
      fprintf (stderr, "%s: Illegal option `%s'.\n", *argv, *parameters);
      goto Usage;
    }
  }

  if (prinit ()) {
    fprintf (stderr, "%s: Could not get the I/O permissions.\n", *argv);
    return 4;
  }

  send_switch ();
  wait_output (REQ_INFO); /* request machine type info */
  receive_switch ();

  if (20 != wait_input ()) {
    fprintf (stderr, "%s: The server is not a VIC-20.\n", *argv);
    input (); input (); input (); input (); /* read the rest of the info */
    return 5;
  }

  driveraddr = input();
  driveraddr |= input() << 8;
  basicaddr = input();
  basicaddr |= input() << 8;

  fprintf (stderr, "%s: Driver address %04X, BASIC start address %04X.\n",
                   *argv, driveraddr, basicaddr);

  if (*parameters) file = NULL;

  while (*parameters || (file == stdin && !feof (file))) {
    unsigned char *buf;
    buf = buffer;

    if (file != stdin && !(file = fopen(*parameters, "rb"))) {
      fprintf (stderr, "%s: Could not open the file `%s'.\n", *argv,
                       *parameters);
      return 5;
    }

    length = fread (buffer, 1, 8192, file);

    if ((buf[0] | (buf[1] << 8)) == start) { /* start address included */
      length -= 2;
      buf += 2;

      if (!feof (file)) /* read the remaining bytes */
	length += fread (&buf[length], 1, 2, file);
    }
    else if (start == 0x6000) { /* alternate start addresses */
      start = buf[0] | (buf[1] << 8);

      if (start == 0x2000 || start == 0x4000) {
        length -= 2;
        buf += 2;

        if (!feof (file)) /* read the remaining bytes */
        length += fread (&buf[length], 1, 2, file);
      }
      else
        start = 0x6000;
    }
    else if (length != 8192 && length != 4096) {
    InvalidFile:
      fprintf (stderr, "%s: File '%s' is not a cartridge image.\n", *argv,
	       file == stdin ? "-" : *parameters);

      if (file != stdin) fclose (file);
      return 5;
    }

    if (file != stdin) {
      fgetc (file);

      if (!feof (file))
        goto InvalidFile;
    }

    if (start == 0xa000 && !flag &&
        !memcmp (&buf[4], "A0\xC3\xC2\xCD", 5)) {
      /* a0CBM autostart code found and no other jumpaddress specified */
      flag = 2;

      auxstart[0] = buf[0];
      auxstart[1] = buf[1];
      buf[0] = (unsigned char)AUXJUMP;
      buf[1] = AUXJUMP >> 8;
    }

    send_switch ();
    output (REQ_LOAD); /* op code for upload */
    output (0); /* specify the bank address */
    receive_switch ();

    if (input ()) {
    notReady:
      fprintf (stderr, "%s: not ready to transfer\n", *argv);
      fclose (file);
      return 6;
    }
    else
      fprintf (stderr, "%s: ok to transfer\n", *argv);

    send_switch ();
    output ((unsigned char)start);
    output ((unsigned char)(start >> 8));
    output ((unsigned char)(start + length));
    output ((unsigned char)((start + length) >> 8));

    psend (buf, length);

    if (file == stdin) {
      unsigned i;

      i = getc (file);

      if (feof (file))
        break;

      ungetc (i, file);
    }
    else {
      fclose (file);
      parameters++;
    }

    if (start == 0xa000)
      start = 0x6000;
    else
      break;
  }

  if (flag) { /* start the cartridge */
    if (flag == 2) {
      send_switch ();
      output (REQ_LOAD); /* first send the startup code */
      output (0);
      receive_switch ();

      if (input ())
        goto notReady;

      fprintf (stderr, "%s: ok to transfer\n", *argv);

      send_switch ();
      output ((unsigned char)auxaddr);
      output ((unsigned char)(auxaddr >> 8));
      output ((unsigned char)(auxaddr + sizeof auxcode));
      output ((unsigned char)((auxaddr + sizeof auxcode) >> 8));

      srvstart[0] = (unsigned char)driveraddr;
      srvstart[1] = driveraddr >> 8;

      psend (auxcode, sizeof auxcode);

      jumpaddress = auxaddr;

      fprintf (stderr, "%s: Type SYS%u to restart the game after reset.\n",
                       *argv, auxaddr);
      fprintf (stderr, "%s: If the game does not start properly, try SYS%u.\n",
		       *argv, ALTJUMP);
    }

    send_switch ();
    output (REQ_JUMP);
    output (0);
    output ((unsigned char)jumpaddress);
    output ((unsigned char)(jumpaddress >> 8));
  }

  prclose ();

  return 0;
}
