/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define BUFSIZE 256
#include "prtrans.h"
#include "prcommon.h"

#if !defined(__MAIN_C__)
#define main	main_prdisk
#endif

#define SECTORS     1

int main(int argc, char **argv);

int
main(int argc, char **argv)
{
  FILE *file;
  unsigned char device = 8, byte;
  int i, writeflag = FALSE;
  char **parameters = argv;
  char *localfile = NULL;
  unsigned driveraddr;
  static char prgname[128];
  char *prg = prgname; /* loader program file name */
  int sectorcount = 0;

  baseaddr = portaddr[DEFAULT_PORT];

  while (*++parameters && **parameters == '-') { /* check for options */
    if (parameters[0][1] == '-') { /* "--" ends options */
      parameters++;
      break;
    }

    switch (parameters[0][1]) {
    case 'd':
      device = strtoul (*++parameters, NULL, 0);

      if (device > 15) {
	fprintf (stderr, "%s: Illegal device number specified.\n", *argv);
	return 1;
      }

      break;

    case 'r':
      writeflag = FALSE;
      break;
    case 'w':
      writeflag = TRUE;
      break;

    case 'l':
      prg = *++parameters;
      break;

    case 'p':
      baseaddr = strtoul (*++parameters, NULL, 16);

      if (baseaddr > 3) {
	fprintf (stderr, "%s: The printer port number must be between 0 and 3.\n",
			 *argv);
	return 1;
      }

      baseaddr = portaddr[baseaddr];
      break;

    case '?':
    case 'h':
    Usage:
      fprintf (stderr, "%s: Copies a file to or from a disk accessible "
		       " from a remote computer.\n\n", *argv);
      fprintf (stderr, "Usage: %s [options] [filename]\n",
		       *argv);
      fprintf (stderr, "Options:\n\t"
			 "-d device\t"
			     "Specify the device number (0-15)\n\t"
			 "-r\t"
			     "Specify read operation (default)\n\t"
			 "-w\t"
			     "Specify write operation\n\t"
			 "-l filename\t"
			     "Specify the filename of the client program\n\t"
			 "-p port\t"
			     "Specify the printer port (0 to 3)\n");
      return 1;

    default:
      fprintf (stderr, "%s: Illegal option `%s'.\n", *argv, *parameters);
      goto Usage;
    }
  }

  if (prinit()) {
    fprintf (stderr, "%s: Could not get the I/O permissions.\n", *argv);
    return 1;
  }

  /* determine the program name to download */
  send_switch ();
  wait_output (REQ_INFO);
  receive_switch ();
  sprintf (prgname, "PRLINK_PRDISK%u", wait_input());
  getname (prgname, prgname, "prdisk.prg");
  driveraddr = input();
  driveraddr |= input() << 8;
  input(); input(); /* skip the BASIC address information */

  if (*parameters) {
    localfile = *parameters++;

    if (!(file = fopen(localfile, writeflag ? "rb" : "wb"))) {
      fprintf (stderr, "%s: Could not open the file `%s'.\n", *argv,
	       localfile);
      prclose();
      return 1;
    }
  }
  else
    file = writeflag ? stdin : stdout;

  {
    FILE *f = fopen (prg, "r");
    if (!f) {
      fprintf (stderr, "%s: Cannot find %s to download\n", *argv, prg);
      prclose();
      return 1;
    }

    if (prloadrun (driveraddr, f)) {
      fprintf (stderr, "%s: Could not upload and start %s\n", *argv, prg);
      prclose();
      return 6;
    }

    fclose(f);
  }

  receive_switch ();

  if (wait_input ()) {
    fprintf (stderr, "%s: not ready to transfer\n", *argv);
    prclose();
    return 6;
  }
  else
    fprintf (stderr, "%s: ok to transfer\n", *argv);

  send_switch ();
  /* specify function: write or read a disk */
  output ((unsigned char)(writeflag ? 2 : 1));
  output (device); /* specify device number */
  receive_switch ();

  if (wait_input ()) {
    fprintf (stderr, "%s: I/O error on remote side\n", *argv);
    fclose (file);
    send_switch ();
    output (0);
    prclose ();
    return 7;
  }

  if (writeflag) { /* loop to write all disk blocks */
    while (!feof (file)) {

      sectorcount++;
      if (sectorcount % SECTORS == 0) {
	fprintf(stderr, " sector %d\r", sectorcount);
	fflush(stderr);
      }

      if (256 != fread (buffer + 1, 1, 256, file)) {
	fprintf (stderr, "\n%s: Could not read all of the input file.\n",
		 *argv);
	break;
      }
      /* calculate the checksum */
      for (*buffer = i = 0; i < 256; *buffer += buffer[++i]);

      send_switch ();
      output (0); /* notify start of block */

      for (;;) {
	psend (buffer, 257);

	receive_switch ();
	byte = wait_input ();

	if (byte != 0x81)
	  break;

	fprintf(stderr, "\n%s: checksum error, retrying\n", *argv);
	send_switch ();
      }

      if (byte != 0x80) {
	fprintf (stderr, "\n%s: remote side aborted the transmission "
			 "with error code %u.\n", *argv, input ());
	goto abort;
      }
    }

    send_switch ();
    output (1); /* end the transfer */
  }
  else { /* loop to get all disk blocks */
    for (;;) {
      for (;;) {
	int chksum;

	sectorcount++;
	if (sectorcount % SECTORS == 0) {
	  fprintf(stderr, " sector %d\r", sectorcount);
	  fflush(stderr);
	}

	receive_switch ();
	if ((byte = wait_input ())) {
	  if (byte == 66) /* illegal track or sector (end of disk) */
	    break;

	  fprintf (stderr, "\n%s: disk error %u\n", *argv, byte);
	abort:
	  fclose (file);
	  send_switch ();
	  output (0);	/* terminate remote side */
	  prclose ();
	  return 7;
	}

	receive_switch ();
	preceive(buffer, 257);
	send_switch ();

	for (chksum = i = 0; i < 256; chksum += buffer[++i]);

	if ((chksum & 0xFF) == buffer[0]) {
	  output(0x80);
	  break;
	}
	fprintf(stderr, "\nchecksum error "
			"(received %02x, calculated %02x), retrying\n",
		buffer[0], (chksum & 0xFF));
	output(0x81);
      }

      if (byte == 66)
	break; /* end of disk */

      fwrite (buffer + 1, 1, 256, file);
    }
  }

  fprintf(stderr, "\n");
  fclose (file);

  send_switch ();
  output (0);	/* terminate remote side */
  prclose ();
  return 0;
}
