/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Main program that combines the previously separate programs prload,
 * prsave, prcart, prrfile, prwfile, prdisk.
 * 
 * Each of the components can be selected by renaming or linking the
 * executable to one of the names, or by supplying an initial parameter of
 * the form --name.
 */

#include <stdio.h>
#include <string.h>

#define __MAIN_C__
#define __PRMAIN_C__
#define BUFSIZE 65536
#include "prtrans.h"

struct program {
  char   *name;
  main_t *main;
};

struct program  plist[] = {
  { "prcart",	&main_prcart },
  { "prdisk",   &main_prdisk },
  { "prload",	&main_prload },
  { "prrfile",	&main_prrfile },
  { "prsave",	&main_prsave },
  { "prwfile",	&main_prwfile },
  { NULL, NULL }
};

struct program *p;

#undef P
#define P(func) func

struct cable {
  char   *name;
  int (*prinit) (void);
  void (*prclose) (void);
  unsigned (*input) (void);
  void (*output) (unsigned char byte);
#ifndef MSDOS
  unsigned (*wait_input) (void);
  void (*wait_output) (unsigned char byte);
#endif
  void (*send_switch) (void);
  void (*receive_switch) (void);
  void (*preceive) (unsigned char *buffer, unsigned length);
  void (*psend) (unsigned char *buffer, unsigned length);
};

struct cable *c;

#undef E
#undef P
#define E extern
#ifdef __amiga__
#  define P(func) func ## _pramiga88
#  include "prtransd.h"
#  undef P
#  define P(func) func ## _pramiga48
#  include "prtransd.h"
#  undef P
#  define P(func) func ## _pramigatn
#  include "prtransd.h"
#  undef P
#  define P(func) func ## _pramieci
#  include "prtransd.h"
#  undef P
#  define P(func) func ## _pramiec
#  include "prtransd.h"
#else
#  define P(func) func ## _prtrans9
#  include "prtransd.h"
#  undef P
#  define P(func) func ## _prtrans8
#  include "prtransd.h"
#  undef P
#  define P(func) func ## _prtrans4
#  include "prtransd.h"
#  undef P
#  define P(func) func ## _prtrans2
#  include "prtransd.h"
#  undef P
#  define P(func) func ## _prtrans1
#  include "prtransd.h"
#endif
#undef P
#undef E
#define P(func) func

#ifndef MSDOS
#define CABLE(s) { #s, \
		   &prinit_ ## s, &prclose_ ## s, &input_ ## s, &output_ ##s, \
		   &wait_input_ ## s, &wait_output_ ## s, \
		   &send_switch_ ## s, &receive_switch_ ## s, \
		   &preceive_ ## s, &psend_ ## s }
#else
#define CABLE(s) { #s, \
		   &prinit_ ## s, &prclose_ ## s, &input_ ## s, &output_ ##s, \
		   &send_switch_ ## s, &receive_switch_ ## s, \
		   &preceive_ ## s, &psend_ ## s }
#endif

struct cable clist[] = {
#ifdef __amiga__
  CABLE(pramiga88),
  CABLE(pramiga48),
  CABLE(pramigatn),
  CABLE(pramiec),
  CABLE(pramieci),
#else
  CABLE(prtrans9),
  CABLE(prtrans8),
  CABLE(prtrans4),
  CABLE(prtrans2),
  CABLE(prtrans1),
#endif
  { NULL, }
};

int findmain (int argc, char **argv)
{
  char *myname = *argv;

  for (p = plist; p->name; p++)
#if 1
    if (strstr(p->name, myname))
      /* Name contains one of our "main" names */
#else
    if (strcmp(p->name, myname) == 0) 
      /* Name matches, call appropriate main_ function */
#endif
      return (*p->main)(argc, argv);

  p = NULL;
  return -1;
}

int findcable (char *name)
{
  for (c = clist; c->name; c++)
#if 1
    if (strstr(c->name, name)) {
#else
    if (strcmp(c->name, name) == 0) {
#endif
      prinit = c->prinit;
      prclose = c->prclose;
      input = c->input;
      output = c->output;
      preceive = c->preceive;
      psend = c->psend;
      send_switch = c->send_switch;
      receive_switch = c->receive_switch;
#ifndef MSDOS
      wait_input = c->wait_input;
      wait_output = c->wait_output;
#endif
      return 0;
    }

  return -1;
}

int main(int argc, char **argv)
{
  char *myname = *argv;
  int exitcode = -1;

  if (findcable (clist->name)) {
    fprintf(stderr,
	    "%s: The protocol could not be set.\n", myname);
    return -1;
  }

  if (argc > 1 && !strcmp(argv[1], "-c")) { /* -c option: specify the cable */
    if (argc < 3 || findcable(argv[2])) {
      fprintf(stderr,
	      "%s: Invalid cable specified.  Available options are:\n",
	      myname);
      for (c = clist; c->name; c++)
	fprintf(stderr, "\t%s\n", c->name);

      return -1;
    }
    argv[2] = *argv;
    argc -= 2;
    argv += 2;
  }

  /* First attempt, try first argument */
  if (argc > 1 && argv[1][0] == '-' && argv[1][1] == '-') {
    /* Adjust arguments */
    int ac = argc - 1;
    char **av = &argv[1];

    av[0] += 2;
    exitcode = findmain(ac, av);
    av[0] -= 2;
  }

  if (p == NULL) {
    /* Second, check the name by which we were invoked */
    exitcode = findmain(argc, argv);
  }

  if (p == NULL) {
    /* Still no go, thow up hands in despair */
    fprintf(stderr,
	    "%s: must specify which function to fulfill, either by renaming the\n"
	    "executable or specifying a first argument of --name.\n"
	    "Available functions are:\n",
	    myname);
    for (p = plist; p->name; p++)
      fprintf(stderr, "\t%s\n", p->name);
  }

  return exitcode;
}
